"""Module to load and init namespace at package level"""
from .wikidata import Wikidata  # noqa: F401
from .chembl import ChEMBL  # noqa: F401
from .ncit import NCIt  # noqa: F401
from .drugbank import DrugBank  # noqa: F401
from .chemidplus import ChemIDplus  # noqa: F401
from .rxnorm import RxNorm  # noqa: F401
from .hemonc import HemOnc  # noqa: F401
from .merge import Merge  # noqa: F401
